#' @keywords internal
#' @useDynLib ragg, .registration = TRUE
#' @importFrom systemfonts match_font
#' @importFrom textshaping text_width
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL
